@@ -20,6 +20,7 @@
 
 #define _POSIX_C_SOURCE 200809L
 #define _BSD_SOURCE
+#define _DEFAULT_SOURCE
 
 #include <glib.h>
 #include <errno.h>
@@ -32,6 +33,8 @@
 #include <unistd.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <time.h>
+#include <gio/gio.h>
 
 
 #include "handle_internal.h"
@@ -56,17 +59,35 @@ lr_get_curl_handle()
     lr_global_init();
 
     h = curl_easy_init();
-    curl_easy_setopt(h, CURLOPT_FOLLOWLOCATION, 1);
-    curl_easy_setopt(h, CURLOPT_MAXREDIRS, 6);
-    curl_easy_setopt(h, CURLOPT_CONNECTTIMEOUT, LRO_CONNECTTIMEOUT_DEFAULT);
-    curl_easy_setopt(h, CURLOPT_LOW_SPEED_TIME, LRO_LOWSPEEDTIME_DEFAULT);
-    curl_easy_setopt(h, CURLOPT_LOW_SPEED_LIMIT, LRO_LOWSPEEDLIMIT_DEFAULT);
-    curl_easy_setopt(h, CURLOPT_SSL_VERIFYHOST, 2);
-    curl_easy_setopt(h, CURLOPT_SSL_VERIFYPEER, 1);
-    curl_easy_setopt(h, CURLOPT_FTP_USE_EPSV, LRO_FTPUSEEPSV_DEFAULT);
-    curl_easy_setopt(h, CURLOPT_FILETIME, 0);
+    if (!h)
+        return NULL;
+
+    if (curl_easy_setopt(h, CURLOPT_FOLLOWLOCATION, 1) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_MAXREDIRS, 6) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_CONNECTTIMEOUT, LRO_CONNECTTIMEOUT_DEFAULT) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_LOW_SPEED_TIME, LRO_LOWSPEEDTIME_DEFAULT) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_LOW_SPEED_LIMIT, LRO_LOWSPEEDLIMIT_DEFAULT) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_SSL_VERIFYHOST, 2) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_SSL_VERIFYPEER, 1) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_SSL_VERIFYSTATUS, 0) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_FTP_USE_EPSV, LRO_FTPUSEEPSV_DEFAULT) != CURLE_OK)
+        goto err;
+    if (curl_easy_setopt(h, CURLOPT_FILETIME, 0) != CURLE_OK)
+        goto err;
 
     return h;
+
+err:
+    curl_easy_cleanup(h);
+    return NULL;
 }
 
 void
@@ -102,6 +123,9 @@ lr_handle_init(void)
     handle->lowspeedlimit = LRO_LOWSPEEDLIMIT_DEFAULT;
     handle->sslverifypeer = 1;
     handle->sslverifyhost = 2;
+    handle->sslverifystatus = 0;
+    handle->proxy_sslverifypeer = 1;
+    handle->proxy_sslverifyhost = 2;
     handle->ipresolve = LRO_IPRESOLVE_DEFAULT;
     handle->allowed_mirror_failures = LRO_ALLOWEDMIRRORFAILURES_DEFAULT;
     handle->adaptivemirrorsorting = LRO_ADAPTIVEMIRRORSORTING_DEFAULT;
@@ -140,6 +164,9 @@ lr_handle_free(LrHandle *handle)
     lr_free(handle->sslclientcert);
     lr_free(handle->sslclientkey);
     lr_free(handle->sslcacert);
+    lr_free(handle->proxy_sslclientcert);
+    lr_free(handle->proxy_sslclientkey);
+    lr_free(handle->proxy_sslcacert);
     lr_lrmirrorlist_free(handle->internal_mirrorlist);
     lr_lrmirrorlist_free(handle->urls_mirrors);
     lr_lrmirrorlist_free(handle->mirrorlist_mirrors);
@@ -240,6 +267,51 @@ lr_handle_remote_sources_changed(LrHandle *handle, LrChangedRemoteSource type)
     }
 }
 
+struct callback_data {
+    GMainLoop *loop; 
+    guint64 deadline_millis;
+    guint timeout_id; 
+    GNetworkMonitor *monitor;
+    GSocketConnectable *connectable; 
+    GCancellable *cancellable;
+};
+
+gboolean 
+timeout_callback(gpointer data)
+{ 
+    struct callback_data *dt = (struct callback_data*)data;
+
+    //if past deadline, exit
+    if (g_get_monotonic_time() >= dt->deadline_millis) {
+        g_main_loop_quit(dt->loop);
+        return G_SOURCE_REMOVE;
+    }
+
+    //if no internet, remove source
+    if (!g_network_monitor_get_network_available(dt->monitor)) {
+        return G_SOURCE_REMOVE;
+    }
+  
+    //if url reached, quit loop and remove source
+    if (g_network_monitor_can_reach (dt->monitor, dt->connectable, dt->cancellable, NULL)) {
+        g_main_loop_quit(dt->loop);
+        return G_SOURCE_REMOVE;
+    }
+
+    //Still waiting on the URL to be availbale, keep polling
+    return G_SOURCE_CONTINUE;
+}
+
+void 
+on_network_available(GObject *object, GParamSpec *pspec, gpointer data)
+{
+    struct callback_data *dt = (struct callback_data*)data;
+    if (dt->timeout_id == 0 && g_network_monitor_get_network_available(dt->monitor)) {
+        dt->timeout_id = g_timeout_add(200, timeout_callback, dt);
+    }
+    g_main_loop_run(dt->loop);
+}
+
 gboolean
 lr_handle_setopt(LrHandle *handle,
                  GError **err,
@@ -604,7 +676,7 @@ lr_handle_setopt(LrHandle *handle,
                         "Value of LRO_LOWSPEEDTIME is too low.");
             ret = FALSE;
         } else {
-            curl_easy_setopt(c_h, CURLOPT_LOW_SPEED_TIME, val_long);
+            c_rc = curl_easy_setopt(c_h, CURLOPT_LOW_SPEED_TIME, val_long);
             handle->lowspeedtime = val_long;
         }
 
@@ -624,7 +696,7 @@ lr_handle_setopt(LrHandle *handle,
                         val_long, handle->maxspeed);
             ret = FALSE;
         } else {
-            curl_easy_setopt(c_h, CURLOPT_LOW_SPEED_LIMIT, val_long);
+            c_rc = curl_easy_setopt(c_h, CURLOPT_LOW_SPEED_LIMIT, val_long);
             handle->lowspeedlimit = val_long;
         }
 
@@ -639,6 +711,11 @@ lr_handle_setopt(LrHandle *handle,
         c_rc = curl_easy_setopt(c_h, CURLOPT_SSL_VERIFYPEER, handle->sslverifypeer);
         break;
 
+    case LRO_SSLVERIFYSTATUS:
+        handle->sslverifystatus = va_arg(arg, long) ? 1 : 0;
+        c_rc = curl_easy_setopt(c_h, CURLOPT_SSL_VERIFYSTATUS, handle->sslverifystatus);
+        break;
+
     case LRO_SSLVERIFYHOST:
         handle->sslverifyhost = va_arg(arg, long) ? 2 : 0;
         c_rc = curl_easy_setopt(c_h, CURLOPT_SSL_VERIFYHOST, handle->sslverifyhost);
@@ -649,6 +726,9 @@ lr_handle_setopt(LrHandle *handle,
             lr_free(handle->sslclientcert);
         handle->sslclientcert = g_strdup(va_arg(arg, char *));
         c_rc = curl_easy_setopt(c_h, CURLOPT_SSLCERT, handle->sslclientcert);
+        if (c_rc == CURLE_OK && handle->sslclientcert && !strncasecmp(handle->sslclientcert, "pkcs11:", 7)) {
+            c_rc = curl_easy_setopt(c_h, CURLOPT_SSLCERTTYPE, "ENG");
+        }
         break;
 
     case LRO_SSLCLIENTKEY:
@@ -656,6 +736,9 @@ lr_handle_setopt(LrHandle *handle,
             lr_free(handle->sslclientkey);
         handle->sslclientkey = g_strdup(va_arg(arg, char *));
         c_rc = curl_easy_setopt(c_h, CURLOPT_SSLKEY, handle->sslclientkey);
+        if (c_rc == CURLE_OK && handle->sslclientkey && !strncasecmp(handle->sslclientkey, "pkcs11:", 7)) {
+            c_rc = curl_easy_setopt(c_h, CURLOPT_SSLKEYTYPE, "ENG");
+        }
         break;
 
     case LRO_SSLCACERT:
@@ -665,6 +748,43 @@ lr_handle_setopt(LrHandle *handle,
         c_rc = curl_easy_setopt(c_h, CURLOPT_CAINFO, handle->sslcacert);
         break;
 
+    case LRO_PROXY_SSLVERIFYPEER:
+        handle->proxy_sslverifypeer = va_arg(arg, long) ? 1 : 0;
+        c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_SSL_VERIFYPEER, handle->proxy_sslverifypeer);
+        break;
+
+    case LRO_PROXY_SSLVERIFYHOST:
+        handle->proxy_sslverifyhost = va_arg(arg, long) ? 2 : 0;
+        c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_SSL_VERIFYHOST, handle->proxy_sslverifyhost);
+        break;
+
+    case LRO_PROXY_SSLCLIENTCERT:
+        if (handle->proxy_sslclientcert)
+            lr_free(handle->proxy_sslclientcert);
+        handle->proxy_sslclientcert = g_strdup(va_arg(arg, char *));
+        c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_SSLCERT, handle->proxy_sslclientcert);
+        if (c_rc == CURLE_OK && handle->proxy_sslclientcert && !strncasecmp(handle->proxy_sslclientcert, "pkcs11:", 7)) {
+            c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_SSLCERTTYPE, "ENG");
+        }
+        break;
+
+    case LRO_PROXY_SSLCLIENTKEY:
+        if (handle->proxy_sslclientkey)
+            lr_free(handle->proxy_sslclientkey);
+        handle->proxy_sslclientkey = g_strdup(va_arg(arg, char *));
+        c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_SSLKEY, handle->proxy_sslclientkey);
+        if (c_rc == CURLE_OK && handle->proxy_sslclientkey && !strncasecmp(handle->proxy_sslclientkey, "pkcs11:", 7)) {
+            c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_SSLKEYTYPE, "ENG");
+        }
+        break;
+
+    case LRO_PROXY_SSLCACERT:
+        if (handle->proxy_sslcacert)
+            lr_free(handle->proxy_sslcacert);
+        handle->proxy_sslcacert = g_strdup(va_arg(arg, char *));
+        c_rc = curl_easy_setopt(c_h, CURLOPT_PROXY_CAINFO, handle->proxy_sslcacert);
+        break;
+
     case LRO_IPRESOLVE: {
         long type = -1;
         long lr_type = va_arg(arg, LrIpResolveType);
@@ -769,6 +889,60 @@ lr_handle_setopt(LrHandle *handle,
     return ret;
 }
 
+gboolean
+lr_handle_network_wait(LrHandle *handle, GError **err, guint seconds, GCancellable *cancellable)
+{
+    assert(!err || *err == NULL);
+
+    if (!handle) {
+        g_set_error(err, LR_HANDLE_ERROR, LRE_BADFUNCARG,
+                    "No handle specified");
+        return FALSE;
+    }
+
+    GNetworkMonitor *monitor = g_network_monitor_get_default();
+
+    struct callback_data data_struct;
+    data_struct.cancellable = cancellable;
+    data_struct.monitor = monitor;
+
+    const gchar *baseurl = NULL;
+    if (handle->metalinkurl)
+        baseurl = handle->metalinkurl;
+    else if (handle->mirrorlisturl)
+        baseurl = handle->mirrorlisturl;
+    else if (handle->urls)
+        baseurl = handle->urls[0];
+    assert(baseurl);
+
+    g_autoptr(GUri) uri = g_uri_parse(baseurl, G_URI_FLAGS_NONE, NULL);
+    if (uri == NULL) {
+        return FALSE;
+    }
+    const gchar* scheme = g_uri_get_scheme(uri);
+    if (!g_strcmp0(scheme, "file")) {
+        return TRUE; 
+    }
+    const gchar* host = g_uri_get_host(uri);
+    guint16 port = g_uri_get_port(uri);
+    GSocketConnectable *connectable =  g_network_address_new(host, port);
+    data_struct.connectable = connectable;
+    data_struct.deadline_millis = g_get_monotonic_time() + seconds * G_USEC_PER_SEC;
+    g_autoptr(GMainLoop) loop;
+    loop = g_main_loop_new(NULL, FALSE);
+    data_struct.loop = loop;
+    data_struct.timeout_id = 0; 
+
+    if (g_network_monitor_get_network_available(data_struct.monitor)) {
+        data_struct.timeout_id = g_timeout_add(200, timeout_callback, &data_struct);
+        g_main_loop_run(data_struct.loop);
+    }
+    else{
+        g_signal_connect(monitor, "notify::network-available", G_CALLBACK(on_network_available), &data_struct);    
+    }
+    return TRUE;
+}
+
 static gboolean
 lr_handle_prepare_urls(LrHandle *handle, GError **err)
 {
@@ -813,7 +987,13 @@ lr_yum_download_url_retry(int attempts, LrHandle *lr_handle, const char *url,
 
         g_debug("%s: Attempt #%d to download %s failed: %s",
                 __func__, i, url, (*err)->message);
-        ftruncate(fd, 0);
+
+        if (ftruncate(fd, 0) < 0) {
+            g_set_error(err, LR_DOWNLOADER_ERROR, LRE_IO,
+                        "ftruncate() failed: %s", g_strerror(errno));
+            return FALSE;
+        }
+
         g_clear_error (err);
     }
 }
@@ -833,8 +1013,7 @@ lr_handle_prepare_mirrorlist(LrHandle *handle, gchar *localpath, GError **err)
         return TRUE;
     } else if (localpath && !handle->mirrorlisturl) {
         // Just try to use mirrorlist of the local repository
-        _cleanup_free_ gchar *path;
-        path = lr_pathconcat(localpath, "mirrorlist", NULL);
+        _cleanup_free_ gchar *path = lr_pathconcat(localpath, "mirrorlist", NULL);
 
         if (g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
             g_debug("%s: Local mirrorlist found at %s", __func__, path);
@@ -949,8 +1128,7 @@ lr_handle_prepare_metalink(LrHandle *handle, gchar *localpath, GError **err)
         return TRUE;
     } else if (localpath && !handle->metalinkurl) {
         // Just try to use metalink of the local repository
-        _cleanup_free_ gchar *path;
-        path = lr_pathconcat(localpath, "metalink.xml", NULL);
+        _cleanup_free_ gchar *path = lr_pathconcat(localpath, "metalink.xml", NULL);
 
         if (g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
             g_debug("%s: Local metalink.xml found at %s", __func__, path);
@@ -1484,6 +1662,12 @@ lr_handle_getinfo(LrHandle *handle,
         *lnum = (long) handle->sslverifypeer;
         break;
 
+    case LRI_SSLVERIFYSTATUS:
+        lnum = va_arg(arg, long *);
+        *lnum = (long) handle->sslverifystatus;
+        break;
+
+
     case LRI_SSLVERIFYHOST:
         lnum = va_arg(arg, long *);
         *lnum = (long) (handle->sslverifyhost ? 1 : 0);
@@ -1504,6 +1688,31 @@ lr_handle_getinfo(LrHandle *handle,
         *str = handle->sslcacert;
         break;
 
+    case LRI_PROXY_SSLVERIFYPEER:
+        lnum = va_arg(arg, long *);
+        *lnum = (long) handle->proxy_sslverifypeer;
+        break;
+
+    case LRI_PROXY_SSLVERIFYHOST:
+        lnum = va_arg(arg, long *);
+        *lnum = (long) (handle->proxy_sslverifyhost ? 1 : 0);
+        break;
+
+    case LRI_PROXY_SSLCLIENTCERT:
+        str = va_arg(arg, char **);
+        *str = handle->proxy_sslclientcert;
+        break;
+
+    case LRI_PROXY_SSLCLIENTKEY:
+        str = va_arg(arg, char **);
+        *str = handle->proxy_sslclientkey;
+        break;
+
+    case LRI_PROXY_SSLCACERT:
+        str = va_arg(arg, char **);
+        *str = handle->proxy_sslcacert;
+        break;
+
     case LRI_IPRESOLVE: {
         LrIpResolveType *type = va_arg(arg, LrIpResolveType *);
         *type = (LrIpResolveType) (handle->ipresolve);
